/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Finds a bit that is not used by POSIX in the sa_flags of struct
 * sigaction, and writes to the specified header a definition of the
 * macro SA_UNSUPPORTED as that bit.  We support the case in which not
 * all of the POSIX-defined flags are defined on the current platform,
 * and indeed we do not need this program if they are all defined.
 */

#define _POSIX_C_SOURCE 200809L

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>


int main(int argc, char **argv)
{
    char const *out_header = NULL;
    FILE *fp = NULL;
    int all_flags = 0;
    int unused_bit, shift;

    if (argc != 2) {
        fprintf(stderr, "Usage: %s <out-header>\n", argv[0]);
        return EXIT_FAILURE;
    }

    out_header = argv[1];

#if defined(SA_NOCLDSTOP)
    all_flags |= SA_NOCLDSTOP;
#endif

#if defined(SA_NOCLDWAIT)
    all_flags |= SA_NOCLDWAIT;
#endif

#if defined(SA_NODEFER)
    all_flags |= SA_NODEFER;
#endif

#if defined(SA_ONSTACK)
    all_flags |= SA_ONSTACK;
#endif

#if defined(SA_RESETHAND)
    all_flags |= SA_RESETHAND;
#endif

#if defined(SA_RESTART)
    all_flags |= SA_RESTART;
#endif

#if defined(SA_SIGINFO)
    all_flags |= SA_SIGINFO;
#endif

    unused_bit = INT_MAX;
    for (shift = 1; shift < sizeof(int) * CHAR_BIT; ++shift) {
        unused_bit &= ~ (unused_bit >> shift);
    }

    while (unused_bit & all_flags) {
        unused_bit >>= 1;
    }

    if (! unused_bit) {
        fprintf(stderr, "%s: could not find unused bit", argv[0]);
        return EXIT_FAILURE;
    }

    fp = fopen(out_header, "w");
    if (! fp) {
        perror(out_header);
        return EXIT_FAILURE;
    }

    if (fprintf(
                fp,
                "#ifndef SA_UNSUPPORTED\n#define SA_UNSUPPORTED %#x\n#endif\n",
                unused_bit) < 0) {
        fclose(fp);
        remove(out_header);
        perror(out_header);
        return EXIT_FAILURE;
    }

    if (fclose(fp)) {
        remove(out_header);
        perror(out_header);
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
